"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
//must remain independent from vscode
const PostCss = require("postcss");
const stylable_1 = require("stylable");
const postcss_ast_utils_1 = require("./utils/postcss-ast-utils");
const completion_providers_1 = require("./completion-providers");
const selector_analyzer_1 = require("./utils/selector-analyzer");
const ts = require("typescript");
const uri_utils_1 = require("./utils/uri-utils");
const lodash_1 = require("lodash");
const pseudo_class_1 = require("./feature/pseudo-class");
const path = require("path");
const pvp = require('postcss-value-parser');
const psp = require('postcss-selector-parser');
const cst = require('css-selector-tokenizer');
class Provider {
    constructor(styl, tsLangService) {
        this.styl = styl;
        this.tsLangService = tsLangService;
        this.providers = [
            completion_providers_1.RulesetInternalDirectivesProvider,
            completion_providers_1.ImportInternalDirectivesProvider,
            completion_providers_1.TopLevelDirectiveProvider,
            completion_providers_1.ValueDirectiveProvider,
            completion_providers_1.GlobalCompletionProvider,
            completion_providers_1.SelectorCompletionProvider,
            completion_providers_1.ExtendCompletionProvider,
            completion_providers_1.CssMixinCompletionProvider,
            completion_providers_1.CodeMixinCompletionProvider,
            completion_providers_1.FormatterCompletionProvider,
            completion_providers_1.NamedCompletionProvider,
            completion_providers_1.StateTypeCompletionProvider,
            completion_providers_1.StateSelectorCompletionProvider,
            completion_providers_1.StateEnumCompletionProvider,
            completion_providers_1.PseudoElementCompletionProvider,
            completion_providers_1.ValueCompletionProvider,
        ];
    }
    provideCompletionItemsFromSrc(src, pos, fileName, fs) {
        let res = fixAndProcess(src, pos, fileName);
        const completions = [];
        try {
            let options = this.createProviderOptions(src, pos, res.processed.meta, res.processed.fakes, res.currentLine, res.cursorLineIndex, fs);
            this.providers.forEach(p => {
                completions.push(...p.provide(options));
            });
        }
        catch (e) {
        }
        return this.dedupeComps(completions);
    }
    getDefinitionLocation(src, position, filePath, fs) {
        if (!filePath.endsWith('.st.css')) {
            return Promise.resolve([]);
        }
        ;
        const { word, meta } = getDefSymbol(src, position, filePath, this.styl);
        if (!meta || !word) {
            return Promise.resolve([]);
        }
        ;
        let defs = [];
        let temp = null;
        if (lodash_1.keys(meta.mappedSymbols).find(sym => sym === word.replace('.', ''))) {
            const symb = meta.mappedSymbols[word.replace('.', '')];
            switch (symb._kind) {
                case 'class': {
                    defs.push(new ProviderLocation(meta.source, this.findWord(word.replace('.', ''), fs.get(meta.source).getText(), position)));
                    break;
                }
                case 'var': {
                    defs.push(new ProviderLocation(meta.source, this.findWord(word, fs.get(meta.source).getText(), position)));
                    break;
                }
                case 'import': {
                    let rslvd = null;
                    try {
                        rslvd = this.styl.resolver.resolve(symb);
                    }
                    catch (e) { }
                    let filePath;
                    if (rslvd && rslvd._kind !== 'js') {
                        filePath = rslvd.meta.source;
                    }
                    else {
                        filePath = this.styl.resolvePath(undefined, symb.import.from);
                    }
                    const doc = fs.get(filePath);
                    if (doc.getText() !== '') {
                        defs.push(new ProviderLocation(filePath, this.findWord(word, doc.getText(), position)));
                    }
                    ;
                    break;
                }
            }
        }
        else if (lodash_1.values(meta.mappedSymbols).some(k => {
            if (k._kind === 'class' && lodash_1.keys(k[stylable_1.valueMapping.states]).some(key => key === word)) {
                temp = k;
                return true;
            }
            else {
                return false;
            }
        })) {
            defs.push(new ProviderLocation(meta.source, this.findWord(temp.name, fs.get(meta.source).getText(), position)));
        }
        else if (lodash_1.keys(meta.customSelectors).find(sym => sym === ':--' + word)) {
            defs.push(new ProviderLocation(meta.source, this.findWord(':--' + word, src, position)));
        }
        else if (word.charAt(0) !== word.charAt(0).toLowerCase()) {
            //Default import, link to top of imported stylesheet
            defs.push(new ProviderLocation(meta.source, completion_providers_1.createRange(0, 0, 0, 0)));
        }
        return Promise.resolve(defs);
    }
    getSignatureHelp(src, pos, filePath, fs, paramInfo) {
        if (!filePath.endsWith('.st.css')) {
            return null;
        }
        const { processed: { meta } } = fixAndProcess(src, pos, filePath);
        if (!meta)
            return null;
        const split = src.split('\n');
        const line = split[pos.line];
        let value = '';
        const path = postcss_ast_utils_1.pathFromPosition(meta.rawAst, { line: pos.line + 1, character: pos.character + 1 });
        if (postcss_ast_utils_1.isRoot(lodash_1.last(path))) { // TODO: check your actually on a selector
            return this.getSignatureForStateWithParamSelector(meta, pos, line);
        }
        else if (line.slice(0, pos.character).trim().startsWith(stylable_1.valueMapping.states)) {
            return this.getSignatureForStateWithParamDefinition(meta, pos, line);
        }
        //If last node is not root, we're in a declaration [TODO: or a media query]
        if (line.slice(0, pos.character).trim().startsWith(stylable_1.valueMapping.mixin)) { //TODO: handle multiple lines as well
            value = line.slice(0, pos.character).trim().slice(stylable_1.valueMapping.mixin.length + 1).trim();
        }
        else if (line.slice(0, pos.character).trim().includes(':')) {
            value = line.slice(0, pos.character).trim().slice(line.slice(0, pos.character).trim().indexOf(':') + 1).trim();
        }
        if (/value\(\s*[^\)]*$/.test(value)) {
            return null;
        }
        const parsed = pvp(value);
        let mixin = '';
        const rev = parsed.nodes.reverse()[0];
        if (rev.type === 'function' && !!rev.unclosed) {
            mixin = rev.value;
        }
        else {
            return null;
        }
        ;
        let activeParam = parsed.nodes.reverse()[0].nodes.reduce((acc, cur) => {
            return (cur.type === 'div' ? acc + 1 : acc);
        }, 0);
        if (mixin === 'value') {
            return null;
        }
        if (meta.mappedSymbols[mixin].import.from.endsWith('.ts')) {
            return this.getSignatureForTsModifier(mixin, activeParam, meta.mappedSymbols[mixin].import.from, meta.mappedSymbols[mixin].type === 'default', paramInfo);
        }
        else if (meta.mappedSymbols[mixin].import.from.endsWith('.js')) {
            if (fs.getOpenedFiles().indexOf(uri_utils_1.toVscodePath(meta.mappedSymbols[mixin].import.from.slice(0, -3) + '.d.ts')) !== -1) {
                return this.getSignatureForTsModifier(mixin, activeParam, meta.mappedSymbols[mixin].import.from.slice(0, -3) + '.d.ts', meta.mappedSymbols[mixin].type === 'default', paramInfo);
            }
            else {
                const importPath = meta.mappedSymbols[mixin].import.from;
                const feh = this.styl.resolvePath(undefined, importPath);
                return this.getSignatureForJsModifier(mixin, activeParam, 
                // fs.get(this.resolveImport((meta.mappedSymbols[mixin]! as ImportSymbol).import.from, this.styl, meta)!.source).getText(),
                fs.get(feh).getText(), paramInfo);
            }
        }
        else {
            return null;
        }
    }
    findWord(word, src, position) {
        let split = src.split('\n');
        let regex = '\\b' + '\\.?' + this.escapeRegExp(word.replace('.', '').replace(':--', '')) + '\\b';
        let lineIndex = split.findIndex(l => {
            const reg = RegExp(regex).exec(l);
            return !!reg && l.slice(reg.index - 2, reg.index) !== '::';
        });
        if (lineIndex === -1 || lineIndex === position.line) {
            lineIndex = position.line;
            // lineIndex = split.findIndex(l => l.trim().indexOf(word) !== -1)
        }
        if (lineIndex === -1) {
            return completion_providers_1.createRange(0, 0, 0, 0);
        }
        ;
        let line = split[lineIndex];
        const match = line.match(RegExp(regex));
        if (match) {
            return completion_providers_1.createRange(lineIndex, line.lastIndexOf(word), lineIndex, line.lastIndexOf(word) + word.length);
        }
        else {
            return completion_providers_1.createRange(0, 0, 0, 0);
        }
    }
    escapeRegExp(re) {
        return re.replace(/([.*+?^=!:${}()|\[\]\/\\])/g, "\\$1");
    }
    getSignatureForTsModifier(mixin, activeParam, filePath, isDefault, paramInfo) {
        let sig = extractTsSignature(filePath, mixin, isDefault, this.tsLangService);
        let ptypes = sig.parameters.map(p => {
            return p.name + ":" + p.valueDeclaration.type.getFullText();
        });
        let rtype = sig.declaration.type
            ? sig.declaration.type.getText()
            : "";
        let parameters = sig.parameters.map(pt => {
            let label = pt.name + ":" + pt.valueDeclaration.type.getFullText();
            return paramInfo.create(label);
        });
        let sigInfo = {
            label: mixin + '(' + ptypes.join(', ') + '): ' + rtype,
            parameters
        };
        return {
            activeParameter: activeParam,
            activeSignature: 0,
            signatures: [sigInfo]
        };
    }
    getSignatureForJsModifier(mixin, activeParam, fileSrc, paramInfo) {
        let lines = fileSrc.split('\n');
        let mixinLine = lines.findIndex(l => l.trim().startsWith('exports.' + mixin));
        let docStartLine = lines.slice(0, mixinLine).lastIndexOf(lines.slice(0, mixinLine).reverse().find(l => l.trim().startsWith('/**')));
        let docLines = lines.slice(docStartLine, mixinLine);
        let formattedLines = [];
        docLines.forEach(l => {
            if (l.trim().startsWith('*/')) {
                return;
            }
            if (l.trim().startsWith('/**') && !!l.trim().slice(3).trim()) {
                formattedLines.push(l.trim().slice(3).trim());
            }
            if (l.trim().startsWith('*')) {
                formattedLines.push(l.trim().slice(1).trim());
            }
        });
        const returnStart = formattedLines.findIndex(l => l.startsWith('@returns'));
        const returnEnd = formattedLines.slice(returnStart + 1).findIndex(l => l.startsWith('@')) === -1
            ? formattedLines.length - 1
            : formattedLines.slice(returnStart + 1).findIndex(l => l.startsWith('@')) + returnStart;
        const returnLines = formattedLines.slice(returnStart, returnEnd + 1);
        formattedLines.splice(returnStart, returnLines.length);
        const returnType = /@returns *{(\w+)}/.exec(returnLines[0])
            ? /@returns *{(\w+)}/.exec(returnLines[0])[1]
            : '';
        const summaryStart = formattedLines.findIndex(l => l.startsWith('@summary'));
        let summaryLines = [];
        if (summaryStart !== -1) {
            const summaryEnd = formattedLines.slice(summaryStart + 1).findIndex(l => l.startsWith('@')) === -1
                ? formattedLines.length - 1
                : formattedLines.slice(summaryStart + 1).findIndex(l => l.startsWith('@')) + summaryStart;
            summaryLines = formattedLines.slice(summaryStart, summaryEnd + 1);
            formattedLines.splice(summaryStart, summaryLines.length);
        }
        let params = [];
        while (formattedLines.find(l => l.startsWith('@param'))) {
            const paramStart = formattedLines.findIndex(l => l.startsWith('@param'));
            const paramEnd = formattedLines.slice(paramStart + 1).findIndex(l => l.startsWith('@')) === -1
                ? formattedLines.length - 1
                : formattedLines.slice(paramStart + 1).findIndex(l => l.startsWith('@')) + paramStart;
            const paramLines = formattedLines.slice(paramStart, paramEnd + 1);
            formattedLines.splice(paramStart, paramLines.length);
            if (/@param *{([ \w<>,'"|]*)} *(\w*)(.*)/.exec(paramLines[0])) {
                params.push([
                    /@param *{([ \w<>,'"|]*)} *(\w*)(.*)/.exec(paramLines[0])[1],
                    /@param *{([ \w<>,'"|]*)} *(\w*)(.*)/.exec(paramLines[0])[2],
                    /@param *{([ \w<>,'"|]*)} *(\w*)(.*)/.exec(paramLines[0])[3],
                ]);
            }
        }
        let descLines = [];
        if (formattedLines.find(l => l.startsWith('@description'))) {
            const descStart = formattedLines.findIndex(l => l.startsWith('@description'));
            const descEnd = formattedLines.slice(descStart + 1).findIndex(l => l.startsWith('@')) === -1
                ? formattedLines.length - 1
                : formattedLines.slice(descStart + 1).findIndex(l => l.startsWith('@')) + descStart;
            descLines = formattedLines.slice(descStart, descEnd + 1);
        }
        else if (formattedLines.findIndex(l => l.startsWith('@')) === -1) {
            descLines = formattedLines;
        }
        else {
            descLines = formattedLines.slice(0, formattedLines.findIndex(l => l.startsWith('@')) + 1);
        }
        if (descLines[0] && descLines[0].startsWith('@description')) {
            descLines[0] = descLines[0].slice(12).trim();
        }
        let parameters = params.map(p => paramInfo.create(p[1] + ': ' + p[0], p[2].trim()));
        let sigInfo = {
            label: mixin + '(' + parameters.map(p => p.label).join(', ') + '): ' + returnType,
            documentation: descLines.join('\n'),
            parameters
        };
        return {
            activeParameter: activeParam,
            activeSignature: 0,
            signatures: [sigInfo]
        };
    }
    getSignatureForStateWithParamSelector(meta, pos, line) {
        let word = '';
        const posChar = pos.character + 1;
        const parsed = cst.parse(line);
        if (parsed.nodes[0].type === 'selector') {
            let length = 0;
            parsed.nodes[0].nodes.forEach((node) => {
                if (node.type === 'invalid') {
                    return; // TODO: refactor - handles places outside of a selector
                }
                else if (node.type === 'spacing') {
                    length += node.value.length;
                }
                else {
                    length += node.name.length + 1;
                    if (node.type === 'pseudo-class' && (posChar > length + 1) && (posChar <= length + 2 + node.content.length)) {
                        word = node.name;
                    }
                }
            });
        }
        let stateDef = null;
        if (word) {
            const transformer = new stylable_1.StylableTransformer({
                diagnostics: new stylable_1.Diagnostics(),
                fileProcessor: this.styl.fileProcessor,
                requireModule: () => {
                    throw new Error('Not implemented, why are we here');
                }
            });
            let resolvedElements = transformer.resolveSelectorElements(meta, line);
            resolvedElements[0][0].resolved.forEach(el => {
                const symbolStates = el.symbol[stylable_1.valueMapping.states];
                if (symbolStates && typeof symbolStates[word] === 'object') {
                    stateDef = symbolStates[word];
                }
            });
            if (stateDef) {
                const parameters = pseudo_class_1.resolveStateParams(stateDef);
                const sigInfo = {
                    label: `${word}(${parameters})`,
                    parameters: [{ label: parameters }]
                };
                return {
                    activeParameter: 0,
                    activeSignature: 0,
                    signatures: [sigInfo]
                };
            }
        }
        return null;
    }
    getSignatureForStateWithParamDefinition(meta, pos, line) {
        const res = pseudo_class_1.resolveStateTypeOrValidator(meta, pos, line);
        if (typeof res === 'string') {
            return pseudo_class_1.createStateValidatorSignature(res);
        }
        else if (typeof res === 'boolean') {
            return pseudo_class_1.createStateTypeSignature();
        }
        else {
            return null;
        }
    }
    createProviderOptions(src, position, meta, fakeRules, fullLineText, cursorPosInLine, fs) {
        const transformer = new stylable_1.StylableTransformer({
            diagnostics: new stylable_1.Diagnostics(),
            fileProcessor: this.styl.fileProcessor,
            requireModule: () => {
                throw new Error('Not implemented, why are we here');
            }
        });
        const path = postcss_ast_utils_1.pathFromPosition(meta.rawAst, { line: position.line + 1, character: position.character });
        const astAtCursor = path[path.length - 1];
        const parentAst = astAtCursor.parent ? astAtCursor.parent : undefined;
        const parentSelector = parentAst && postcss_ast_utils_1.isSelector(parentAst) && fakeRules.findIndex((f) => {
            return f.selector === parentAst.selector;
        }) === -1
            ? parentAst
            : astAtCursor && postcss_ast_utils_1.isSelector(astAtCursor) && fakeRules.findIndex((f) => {
                return f.selector === astAtCursor.selector;
            }) === -1
                ? astAtCursor
                : null;
        const { lineChunkAtCursor, fixedCharIndex } = getChunkAtCursor(fullLineText, cursorPosInLine);
        const ps = selector_analyzer_1.parseSelector(lineChunkAtCursor, fixedCharIndex);
        const chunkStrings = ps.selector.reduce((acc, s) => {
            return acc.concat(s.text);
        }, []);
        const currentSelector = ps.selector[0].classes[0] || ps.selector[0].customSelectors[0] || chunkStrings[0];
        const expandedLine = stylable_1.expandCustomSelectors(PostCss.rule({ selector: lineChunkAtCursor }), meta.customSelectors).split(' ').pop(); // TODO: replace with selector parser
        const resolvedElements = transformer.resolveSelectorElements(meta, expandedLine);
        let resolved = [];
        if (currentSelector && resolvedElements[0].length) {
            const clas = resolvedElements[0].find(e => e.type === 'class' || (e.type === 'element' && e.resolved.length > 1)); //TODO: better type parsing
            resolved = clas ? clas.resolved : [];
        }
        return {
            meta: meta,
            fs: fs,
            styl: this.styl,
            src: src,
            tsLangService: this.tsLangService,
            resolvedElements: resolvedElements,
            parentSelector: parentSelector,
            astAtCursor: astAtCursor,
            lineChunkAtCursor: lineChunkAtCursor,
            lastSelectoid: ps.lastSelector,
            fullLineText: fullLineText,
            position: position,
            resolved: resolved,
            currentSelector: currentSelector,
            target: ps.target,
            isMediaQuery: isInMediaQuery(path),
            fakes: fakeRules,
        };
    }
    dedupeComps(completions) {
        let uniqs = new Map();
        completions.forEach(comp => {
            if (!uniqs.has(comp.label)) {
                uniqs.set(comp.label, comp);
            }
        });
        let res = [];
        uniqs.forEach(v => res.push(v));
        return res;
    }
}
exports.default = Provider;
function isIllegalLine(line) {
    return /^\s*[-\.:]+\s*$/.test(line);
}
const lineEndsRegexp = /({|}|;)/;
function findRefs(word, defMeta, curMeta, styl) {
    if (!word) {
        return [];
    }
    const refs = [];
    const trans = styl.createTransformer();
    if (word.startsWith(':global(')) {
        curMeta.rawAst.walkRules(rule => {
            if (rule.selector.includes(word) && rule.source && rule.source.start) {
                refs.push({
                    uri: uri_utils_1.toVscodePath(curMeta.source),
                    range: {
                        start: {
                            line: rule.source.start.line - 1,
                            character: rule.selector.indexOf(word) + ':global('.length
                        },
                        end: {
                            line: rule.source.start.line - 1,
                            character: rule.selector.indexOf(word) + word.length - 1
                        }
                    }
                });
            }
        });
        return refs;
    }
    const valueRegex = new RegExp('(\\.?' + word + ')(\\s|$|\\:|;|\\)|,)', 'g');
    curMeta.rawAst.walkRules((rule) => {
        //Usage in selector
        const filterRegex = new RegExp('(\\.?' + word + ')(\\s|$|\\:|;|\\))', 'g');
        if (filterRegex.test(rule.selector) && !!rule.source && !!rule.source.start) {
            const resLocal = trans.resolveSelectorElements(curMeta, rule.selector);
            if (resLocal[0].some(rl => {
                return rl.name === word && lodash_1.last(rl.resolved).meta.source === defMeta.source;
            })) {
                let match;
                while ((match = valueRegex.exec(rule.selector)) !== null) {
                    const index = match[0].startsWith('.') ? match.index : match.index - 1;
                    refs.push({
                        uri: uri_utils_1.toVscodePath(curMeta.source),
                        range: {
                            start: {
                                line: rule.source.start.line - 1,
                                character: rule.source.start.column + index
                            },
                            end: {
                                line: rule.source.start.line - 1,
                                character: rule.source.start.column + index + word.length
                            }
                        }
                    });
                }
            }
        }
    });
    curMeta.rawAst.walkDecls((decl) => {
        if (!decl.source || !decl.source.start) {
            return;
        }
        const directiveRegex = new RegExp(stylable_1.valueMapping.extends + '|' + stylable_1.valueMapping.named + '|' + stylable_1.valueMapping.default + '|' + stylable_1.valueMapping.states);
        if (directiveRegex.test(decl.prop)) {
            //Usage in -st directives
            const reg = new RegExp(valueRegex.source);
            const match = reg.exec(decl.value);
            if (match) {
                refs.push({
                    uri: uri_utils_1.toVscodePath(curMeta.source),
                    range: {
                        start: {
                            line: decl.source.start.line - 1,
                            character: match.index + decl.source.start.column + decl.prop.length + (decl.raws.between ? decl.raws.between.length : 0) - 1
                        },
                        end: {
                            line: decl.source.start.line - 1,
                            character: match.index + decl.source.start.column + decl.prop.length + (decl.raws.between ? decl.raws.between.length : 0) + word.length - 1
                        }
                    }
                });
            }
        }
    });
    curMeta.rawAst.walkDecls(stylable_1.valueMapping.mixin, (decl) => {
        //usage in mixin
        if (!decl.source || !decl.source.start) {
            return;
        }
        const lines = decl.value.split('\n');
        lines.forEach((line, index) => {
            let match;
            while ((match = valueRegex.exec(line)) !== null) {
                refs.push({
                    uri: uri_utils_1.toVscodePath(curMeta.source),
                    range: {
                        start: {
                            line: decl.source.start.line - 1 + index,
                            character: index
                                ? match.index
                                : decl.source.start.column + stylable_1.valueMapping.mixin.length + match.index + (decl.raws.between ? decl.raws.between.length : 0) - 1
                        },
                        end: {
                            line: decl.source.start.line - 1 + index,
                            character: word.length + (index
                                ? match.index
                                : decl.source.start.column + stylable_1.valueMapping.mixin.length + match.index + (decl.raws.between ? decl.raws.between.length : 0) - 1)
                        }
                    }
                });
            }
        });
    });
    curMeta.rawAst.walkDecls(word, (decl) => {
        //Variable definition
        if (decl.parent.type === 'rule' && decl.parent.selector === ':vars' && !!decl.source && !!decl.source.start) {
            refs.push({
                uri: uri_utils_1.toVscodePath(curMeta.source),
                range: {
                    start: {
                        line: decl.source.start.line - 1,
                        character: decl.source.start.column - 1
                    },
                    end: {
                        line: decl.source.start.line - 1,
                        character: decl.source.start.column + word.length - 1
                    }
                }
            });
        }
    });
    curMeta.rawAst.walkDecls((decl) => {
        //Variable usage
        if (decl.value.includes('value(') && !!decl.source && !!decl.source.start) {
            const usageRegex = new RegExp('value\\(\\s*' + word + '\\s*\\)', 'g');
            const match = usageRegex.exec(decl.value);
            if (match) {
                refs.push({
                    uri: uri_utils_1.toVscodePath(curMeta.source),
                    range: {
                        start: {
                            line: decl.source.start.line - 1,
                            character: match.index + decl.source.start.column + decl.prop.length + (decl.raws.between ? decl.raws.between.length : 0) + 'value('.length - 1
                        },
                        end: {
                            line: decl.source.start.line - 1,
                            character: match.index + decl.source.start.column + decl.prop.length + (decl.raws.between ? decl.raws.between.length : 0) + 'value('.length + word.length - 1
                        }
                    }
                });
            }
        }
    });
    return refs;
}
function newFindRefs(word, meta, files, styl) {
    let refs = [];
    if (word.startsWith(':global(')) { //Global selector strings are special
        files.forEach(file => {
            const newMeta = styl.process(file.fullPath);
            newMeta.rawAst.walkRules(rule => {
                if (rule.selector.includes(word)) {
                    refs = refs.concat(findRefs(word, meta, newMeta, styl));
                }
            });
        });
        return refs;
    }
    else {
        word = word.replace('.', '');
    }
    if (!meta.mappedSymbols[word] && word.charAt(0) !== word.charAt(0).toLowerCase()) { //Default import
        files.forEach(file => {
            const newMeta = styl.process(file.fullPath);
            let tmp = '';
            if (Object.keys(newMeta.mappedSymbols).some(k => {
                tmp = k;
                return ((newMeta.mappedSymbols[k]._kind === 'element'
                    && newMeta.mappedSymbols[k].alias
                    && (newMeta.mappedSymbols[k].alias.import.from === meta.source))
                    || (newMeta.mappedSymbols[k]._kind === 'import'
                        && newMeta.mappedSymbols[k].import.from === meta.source));
            })) {
                refs = refs.concat(findRefs(tmp, meta, newMeta, styl));
            }
        });
    }
    else if (meta.mappedSymbols[word] && meta.mappedSymbols[word]._kind === 'var') { //Vatiable
        files.forEach(file => {
            const newMeta = styl.process(file.fullPath);
            if (!newMeta.mappedSymbols[word] || (newMeta.mappedSymbols[word]._kind !== 'var' && newMeta.mappedSymbols[word]._kind !== 'import')) {
                return;
            }
            if (newMeta.source === meta.source) { //We're in the defining file
                refs = refs.concat(findRefs(word.replace('.', ''), meta, newMeta, styl));
            }
            else { //We're in a using file
                const newSymb = styl.resolver.deepResolve(newMeta.mappedSymbols[word]);
                if (!newSymb || !newSymb.meta) {
                    return;
                }
                if (newSymb.meta.source === meta.source) {
                    refs = refs.concat(findRefs(word.replace('.', ''), meta, newMeta, styl));
                }
            }
        });
    }
    else if (meta.mappedSymbols[word] && (meta.mappedSymbols[word]._kind === 'class' || meta.mappedSymbols[word]._kind === 'import')) { //Elements
        const trans = styl.createTransformer();
        const valueRegex = new RegExp('(\\.?' + word + ')\\b', 'g');
        files.forEach(file => {
            const newMeta = styl.process(file.fullPath);
            let done = false;
            newMeta.rawAst.walkRules(r => {
                if (valueRegex.test(r.selector) && !done) {
                    const resolved = trans.resolveSelectorElements(newMeta, r.selector);
                    const resolvedInner = resolved[0].find(r => r.name === word);
                    if (resolvedInner && resolvedInner.resolved.some(r => r.meta.source === meta.source)) {
                        refs = refs.concat(findRefs(word.replace('.', ''), meta, newMeta, styl));
                        done = true;
                    }
                }
            });
            newMeta.rawAst.walkDecls(d => {
                if (valueRegex.test(d.value) && !done) {
                    if (d.prop === stylable_1.valueMapping.named && d.parent.nodes.find(n => {
                        return n.prop === stylable_1.valueMapping.from
                            && path.resolve(path.dirname(newMeta.source), n.value.replace(/"/g, '')) === meta.source;
                    })) {
                        refs = refs.concat(findRefs(word.replace('.', ''), meta, newMeta, styl));
                        done = true;
                    }
                }
            });
        });
    }
    else if (lodash_1.keys(meta.mappedSymbols).some(k => {
        return meta.mappedSymbols[k]._kind === 'class' && lodash_1.keys(meta.mappedSymbols[k][stylable_1.valueMapping.states]).some(k => k === word);
    })) {
        files.forEach(file => {
            const newMeta = styl.process(file.fullPath);
            let done = false;
            if (meta.source === newMeta.source) {
                refs = refs.concat(findRefs(word.replace('.', ''), meta, newMeta, styl));
                return;
            }
            const trans = styl.createTransformer();
            newMeta.rawAst.walkRules(r => {
                if (r.selector.includes(':' + word) && !done) {
                    const ido = selector_analyzer_1.parseSelector(r.selector, r.selector.indexOf(word)); //Won't work if word appears elsewhere in string
                    const elem = ido.selector[ido.target.index].type === '*'
                        ? ido.selector[ido.target.index].text[0].replace('.', '')
                        : ido.selector[ido.target.index].name;
                    const reso = trans.resolveSelectorElements(newMeta, r.selector);
                    const symb = reso[0].find(o => o.name === elem);
                    if (symb && symb.resolved.some(r => {
                        return (r.symbol._kind === 'class' && r.symbol.name === elem && r.meta.source === meta.source);
                    })) {
                        refs = refs.concat(findRefs(word.replace('.', ''), meta, newMeta, styl));
                        done = true;
                    }
                }
            });
        });
    }
    return refs;
}
function getRefs(params, fs, styl) {
    let refs = [];
    const symb = getDefSymbol(fs.loadTextFileSync(params.textDocument.uri), params.position, params.textDocument.uri, styl);
    if (!symb.meta) {
        return refs;
    }
    ;
    let files = (fs.loadDirectoryTreeSync(styl.projectRoot).children);
    const cont = files.filter(f => f.name.endsWith('.st.css')).map(f => {
        f.content = fs.loadTextFileSync(f.fullPath);
        return f;
    });
    refs = newFindRefs(symb.word, symb.meta, cont, styl);
    return refs;
}
exports.getRefs = getRefs;
function getRenameRefs(params, fs, styl) {
    const refs = getRefs(params, fs, styl);
    const newRefs = [];
    refs.forEach(ref => {
        if (!ref.uri.includes('node_modules') && uri_utils_1.fromVscodePath(ref.uri).startsWith(styl.projectRoot)) {
            newRefs.push(ref);
        }
    });
    return newRefs;
}
exports.getRenameRefs = getRenameRefs;
function createMeta(src, path) {
    let meta;
    let fakes = [];
    try {
        let ast = stylable_1.safeParse(src, { from: uri_utils_1.fromVscodePath(path) });
        ast.nodes && ast.nodes.forEach((node) => {
            if (node.type === 'decl') {
                let r = PostCss.rule({ selector: node.prop + ':' + node.value });
                r.source = node.source;
                node.replaceWith(r);
                fakes.push(r);
            }
        });
        if (ast.raws.after && ast.raws.after.trim()) {
            let r = PostCss.rule({ selector: ast.raws.after.trim() });
            ast.append(r);
            fakes.push(r);
        }
        meta = stylable_1.process(ast);
    }
    catch (error) {
        return { meta: null, fakes: fakes };
    }
    return {
        meta: meta,
        fakes: fakes
    };
}
exports.createMeta = createMeta;
function fixAndProcess(src, position, filePath) {
    let cursorLineIndex = position.character;
    let lines = src.replace(/\r\n/g, '\n').split('\n');
    let currentLine = lines[position.line];
    let fixedSrc = src;
    if (currentLine.match(lineEndsRegexp)) {
        let currentLocation = 0;
        let splitLine = currentLine.split(lineEndsRegexp);
        for (var i = 0; i < splitLine.length; i += 2) {
            currentLocation += splitLine[i].length + 1;
            if (currentLocation >= position.character) {
                currentLine = splitLine[i];
                if (isIllegalLine(currentLine)) {
                    splitLine[i] = '\n';
                    lines.splice(position.line, 1, splitLine.join(''));
                    fixedSrc = lines.join('\n');
                }
                break;
            }
            else {
                cursorLineIndex -= splitLine[i].length + 1;
            }
        }
    }
    else if (isIllegalLine(currentLine)) {
        lines.splice(position.line, 1, "");
        fixedSrc = lines.join('\n');
    }
    let processed = createMeta(fixedSrc, filePath);
    return {
        processed: processed,
        currentLine: currentLine,
        cursorLineIndex: cursorLineIndex,
    };
}
exports.fixAndProcess = fixAndProcess;
class ProviderLocation {
    constructor(uri, range) {
        this.uri = uri;
        this.range = range;
    }
}
exports.ProviderLocation = ProviderLocation;
function extractTsSignature(filePath, mixin, isDefault, tsLangService) {
    tsLangService.setOpenedFiles([filePath]);
    const program = tsLangService.ts.getProgram();
    const tc = program.getTypeChecker();
    const sf = program.getSourceFile(filePath);
    if (!sf) {
        return undefined;
    }
    const mix = tc.getSymbolsInScope(sf, ts.SymbolFlags.Function).find(f => {
        if (isDefault) {
            return f.exportSymbol && f.exportSymbol.escapedName === 'default';
        }
        else {
            return f.exportSymbol && f.exportSymbol.escapedName === mixin;
        }
    });
    if (!mix) {
        return undefined;
    }
    return tc.getSignatureFromDeclaration(mix.declarations[0]);
}
exports.extractTsSignature = extractTsSignature;
function extractJsModifierReturnType(mixin, activeParam, fileSrc) {
    let lines = fileSrc.split('\n');
    let mixinLine = lines.findIndex(l => l.trim().startsWith('exports.' + mixin));
    let docStartLine = lines.slice(0, mixinLine).lastIndexOf(lines.slice(0, mixinLine).reverse().find(l => l.trim().startsWith('/**')));
    let docLines = lines.slice(docStartLine, mixinLine);
    let formattedLines = [];
    docLines.forEach(l => {
        if (l.trim().startsWith('*/')) {
            return;
        }
        if (l.trim().startsWith('/**') && !!l.trim().slice(3).trim()) {
            formattedLines.push(l.trim().slice(3).trim());
        }
        if (l.trim().startsWith('*')) {
            formattedLines.push(l.trim().slice(1).trim());
        }
    });
    const returnStart = formattedLines.findIndex(l => l.startsWith('@returns'));
    const returnEnd = formattedLines.slice(returnStart + 1).findIndex(l => l.startsWith('@')) === -1
        ? formattedLines.length - 1
        : formattedLines.slice(returnStart + 1).findIndex(l => l.startsWith('@')) + returnStart;
    const returnLines = formattedLines.slice(returnStart, returnEnd + 1);
    formattedLines.splice(returnStart, returnLines.length);
    const returnType = /@returns *{(\w+)}/.exec(returnLines[0])
        ? /@returns *{(\w+)}/.exec(returnLines[0])[1]
        : '';
    return returnType;
}
exports.extractJsModifierReturnType = extractJsModifierReturnType;
function isInMediaQuery(path) {
    return path.some(n => n.type === 'atrule' && n.name === 'media');
}
;
function isDirective(line) {
    return lodash_1.keys(stylable_1.valueMapping).some(k => line.trim().startsWith(stylable_1.valueMapping[k]));
}
exports.isDirective = isDirective;
;
function isNamedDirective(line) {
    return line.indexOf(stylable_1.valueMapping.named) !== -1;
}
;
function isInValue(lineText, position) {
    let isInValue = false;
    if (/value\(/.test(lineText)) {
        let line = lineText.slice(0, position.character);
        let stack = 0;
        for (let i = 0; i <= line.length; i++) {
            if (line[i] === '(') {
                stack += 1;
            }
            else if (line[i] === ')') {
                stack -= 1;
            }
        }
        if (stack > 0) {
            isInValue = true;
        }
    }
    return isInValue;
}
exports.isInValue = isInValue;
function getChunkAtCursor(fullLineText, cursorPosInLine) {
    let fixedCharIndex = cursorPosInLine;
    let lineChunkAtCursor = fullLineText;
    while (lineChunkAtCursor.lastIndexOf(' ') >= cursorPosInLine) {
        lineChunkAtCursor = lineChunkAtCursor.slice(0, lineChunkAtCursor.lastIndexOf(' '));
    }
    if (!isDirective(lineChunkAtCursor) && lineChunkAtCursor.lastIndexOf(' ') > -1 && lineChunkAtCursor.lastIndexOf(' ') < cursorPosInLine) {
        fixedCharIndex -= (lineChunkAtCursor.lastIndexOf(' ') + 1);
        lineChunkAtCursor = lineChunkAtCursor.slice(lineChunkAtCursor.lastIndexOf(' '));
    }
    return { lineChunkAtCursor: lineChunkAtCursor.trim(), fixedCharIndex };
}
function getNamedValues(src, lineIndex) {
    let lines = src.split('\n');
    let isNamedValueLine = false;
    let namedValues = [];
    for (let i = lineIndex; i > 0; i--) {
        if (isDirective(lines[i]) && !isNamedDirective(lines[i])) {
            break;
        }
        else if (isNamedDirective(lines[i])) {
            isNamedValueLine = true;
            let valueStart = lines[i].indexOf(':') + 1;
            let value = lines[i].slice(valueStart);
            value.split(',').map(x => x.trim()).filter(x => x !== '').forEach(x => namedValues.push(x));
            break;
        }
        else {
            let valueStart = lines[i].indexOf(':') + 1;
            let value = lines[i].slice(valueStart);
            value.split(',').map(x => x.trim()).filter(x => x !== '').forEach(x => namedValues.push(x));
        }
    }
    return { isNamedValueLine, namedValues };
}
exports.getNamedValues = getNamedValues;
function getExistingNames(lineText, position) {
    const valueStart = lineText.indexOf(':') + 1;
    const value = lineText.slice(valueStart, position.character);
    const parsed = pvp(value.trim());
    const names = parsed.nodes.filter((n) => n.type === 'function' || n.type === 'word').map((n) => n.value);
    const rev = parsed.nodes.reverse();
    const lastName = (parsed.nodes.length && rev[0].type === 'word') ? rev[0].value : '';
    return { names, lastName };
}
exports.getExistingNames = getExistingNames;
function findNode(nodes, index) {
    return nodes
        .filter(n => n.sourceIndex <= index)
        .reduce((m, n) => {
        return (m.sourceIndex > n.sourceIndex) ? m : n;
    }, { sourceIndex: -1 });
}
function getDefSymbol(src, position, filePath, styl) {
    let res = fixAndProcess(src, position, filePath);
    let meta = res.processed.meta;
    if (!meta) {
        return { word: '', meta: null };
    }
    ;
    const parsed = pvp(res.currentLine).nodes;
    let val = findNode(parsed, position.character);
    while (val.nodes && val.nodes.length > 0) {
        if (findNode(val.nodes, position.character).sourceIndex >= 0) {
            val = findNode(val.nodes, position.character);
        }
        else {
            break;
        }
    }
    ;
    let word = val.value;
    const { lineChunkAtCursor } = getChunkAtCursor(res.currentLine.slice(0, val.sourceIndex + val.value.length), position.character);
    const directiveRegex = new RegExp(stylable_1.valueMapping.extends + '|' + stylable_1.valueMapping.named + '|' + stylable_1.valueMapping.default + '|' + stylable_1.valueMapping.mixin);
    if (lineChunkAtCursor.startsWith(':global')) {
        return { word: ':global(' + word + ')', meta };
    }
    const match = lineChunkAtCursor.match(directiveRegex);
    if (match && !!meta.mappedSymbols[word]) {
        // We're in an -st directive
        let imp;
        if (meta.mappedSymbols[word]._kind === 'import') {
            imp = styl.resolver.resolveImport((meta.mappedSymbols[word]));
        }
        else if (meta.mappedSymbols[word]._kind === 'element' && meta.mappedSymbols[word].alias) {
            imp = styl.resolver.resolveImport(meta.mappedSymbols[word].alias);
        }
        else if (meta.mappedSymbols[word]._kind === 'class') {
            if (!!meta.mappedSymbols[word].alias) {
                meta = styl.resolver.resolveImport(meta.mappedSymbols[word].alias).meta;
            }
            return { word, meta };
        }
        if (imp) {
            if (imp._kind === 'js') {
                return { word, meta };
            }
            else
                return { word, meta: imp.meta };
        }
        else {
            return { word: '', meta: null };
        }
    }
    const varRegex = new RegExp('value\\(\\s*' + word);
    if (varRegex.test(lineChunkAtCursor)) {
        // we're looking at a var usage
        if (!meta.mappedSymbols[word]) {
            return { word, meta: null };
        }
        else if (meta.mappedSymbols[word]._kind === 'var') { //deepResolve doesn't do local symbols
            return { word, meta };
        }
        const resolvedVar = styl.resolver.deepResolve(meta.mappedSymbols[word]);
        if (resolvedVar) {
            return { word, meta: resolvedVar.meta };
        }
        else {
            return { word, meta: null };
        }
    }
    const transformer = new stylable_1.StylableTransformer({
        diagnostics: new stylable_1.Diagnostics(),
        fileProcessor: styl.fileProcessor,
        requireModule: () => {
            throw new Error('Not implemented, why are we here');
        }
    });
    const expandedLine = stylable_1.expandCustomSelectors(PostCss.rule({ selector: lineChunkAtCursor }), meta.customSelectors).split(' ').pop(); // TODO: replace with selector parser
    const resolvedElements = transformer.resolveSelectorElements(meta, expandedLine);
    let reso;
    if (word.charAt(0) !== word.charAt(0).toLowerCase()) {
        reso = resolvedElements[0][resolvedElements[0].length - 1].resolved.find(res => !!res.symbol["-st-root"]);
    }
    else {
        reso = resolvedElements[0][resolvedElements[0].length - 1].resolved.find(res => {
            return (res.symbol.name === word.replace('.', '') && !res.symbol.alias) || lodash_1.keys(res.symbol[stylable_1.valueMapping.states]).some(k => k === word);
        });
    }
    if (reso) {
        meta = reso.meta;
    }
    return { word, meta };
}
exports.getDefSymbol = getDefSymbol;
//# sourceMappingURL=provider.js.map